rem Copyright (c) 1996-1998 Mark Lang
rem to add...
rem
rem   FUTURE ENHANCEMENTS???
rem     support COPY/MOVE?/DUMP routines
rem     support SAVE/RESTORE_ALL routines
rem     trim truncate/convert support
rem     take advantage of BIND_ARRAY (v8)
rem     generate DBMS_SQL.BIND_ARRAY code
rem     Partition column update support (v8)
rem     add support for checking user privileges (required DBA_TAB_PRIVS)
rem     check column_name in EXCL... parameters
rem     warn if all non-key columns are filtered out???
rem     relieve 32k limit
rem
rem     RENAME_OLD_VERSION
rem     CREATE_TABLE_LIKE
rem     DROP_COLUMN
rem     DELETE_CASCADE
rem     CREATE / COPY INDEXES
rem     APPLY
rem     APPLY_IN_FK_ORDER        also return in tabtab_t
rem     DESCRIBE
rem     UPDATE_KEY_COLUMN
rem     GLOBAL_UPDATE / UPDATE_COLUMN
rem
rem     generate table triggers to ...
rem     common interval times

rem Notes:
rem if x flag is used the RBSEG is ignored

create or replace package body lang_updtab
as

insufficient_privileges exception;
pragma exception_init(insufficient_privileges, -1031);

-- SHARED TYPES
subtype myint is binary_integer;

-- **************
-- SHARED CURSORS
-- **************


function version
return varchar2 is
begin
  return 'Version 1.0.0 Production - Mar 16, 1998';
end;


function quote(str varchar2)
return varchar2 is
begin
  return ''''||str||'''';
end quote;


function convert_number(str varchar2, fmt varchar2 default null)
return number is
begin
  if fmt is null then
    return to_number(str);
  else
    return to_number(str, fmt);
  end if;
exception
  when value_error then
  return null;
end convert_number;


function convert_date(str varchar2, fmt varchar2 default null)
return date is
begin
  if fmt is null then
    return to_date(str);
  else
    return to_date(str, fmt);
  end if;
exception
  when others then
  return null;
end convert_date;


function concat_if(str1 varchar2, str2 varchar2
, ok boolean default true, sep varchar2 default null)
return varchar2 is
begin
  if ok then
    if str1 is null then
      return str2;
    end if;
    return str1||sep||str2;
  end if;
  return str1;
end concat_if;

-- ******************
-- SHARED SUBROUTINES
-- ******************


procedure out(s varchar2, ok boolean default true)
is
begin
  if ok then
    if length(s) <= 240 then
      sys.dbms_output.put_line(s);
    else
      for i in 1..trunc(length(s) / 240) + 1 loop
	sys.dbms_output.put_line(substr(s, (i-1)*240+1, 240));
      end loop;
    end if;
  end if;
end;


function isflag(flags varchar2, flag_list varchar2) return boolean is
  i binary_integer;
  f varchar2(30);
  l varchar2(30);
  c varchar2(1);
begin
  f := upper(flags);
  l := upper(flag_list);
  for i in 1..length(l) loop
    c := substr(l, i, 1);
    if c != ' ' and not instr(f, c) > 0 then
      return false;
    end if;
  end loop;
  return true;
end isflag;


function orflag(flags varchar2, flag_list varchar2) return boolean is
  i binary_integer;
  f varchar2(30);
  l varchar2(30);
  c varchar2(1);
begin
  f := upper(flags);
  l := upper(flag_list);
  for i in 1..length(l) loop
    c := substr(l, i, 1);
    if c != ' ' and instr(f, c) > 0 then
      return true;
    end if;
  end loop;
  return false;
end orflag;


function exec(stmt varchar2)
return number is
  i number;
  retval number;
begin
  out(stmt, debugging);
  i := sys.dbms_sql.open_cursor;
  sys.dbms_sql.parse(i, stmt, dbms_sql.native);
  retval := sys.dbms_sql.execute(i);
  sys.dbms_sql.close_cursor(i);
  return retval;
end exec;


function execq(stmt varchar2, result in out number)
return number is
  i number;
  retval number;
begin
  i := sys.dbms_sql.open_cursor;
  sys.dbms_sql.parse(i, stmt, dbms_sql.native);
  sys.dbms_sql.define_column(i, 1, result);
  retval := sys.dbms_sql.execute_and_fetch(i);
  if retval > 0 then
    sys.dbms_sql.column_value(i, 1, result);
  end if;
  sys.dbms_sql.close_cursor(i);
  return retval;
end execq;


function execq(stmt varchar2, result in out varchar2, len number default 255)
return number is
  i number;
  retval number;
begin
  i := sys.dbms_sql.open_cursor;
  sys.dbms_sql.parse(i, stmt, dbms_sql.native);
  sys.dbms_sql.define_column(i, 1, result, len);
  retval := sys.dbms_sql.execute_and_fetch(i);
  if retval > 0 then
    sys.dbms_sql.column_value(i, 1, result);
  end if;
  sys.dbms_sql.close_cursor(i);
  return retval;
end execq;


function execq(stmt varchar2, result in out date)
return number is
  i number;
  retval number;
begin
  i := sys.dbms_sql.open_cursor;
  sys.dbms_sql.parse(i, stmt, sys.dbms_sql.native);
  sys.dbms_sql.define_column(i, 1, result);
  retval := sys.dbms_sql.execute_and_fetch(i);
  if retval > 0 then
    sys.dbms_sql.column_value(i, 1, result);
  end if;
  sys.dbms_sql.close_cursor(i);
  return retval;
end execq;


procedure doit(stmt varchar2)
is
  retval number;
begin
  retval := exec(stmt);
end;


procedure doitq(stmt varchar2, result in out number)
is
  retval number;
begin
  retval := execq(stmt, result);
end;


procedure doitq(stmt varchar2, result in out varchar2, len number default 255)
is
  retval number;
begin
  retval := execq(stmt, result, len);
end;


procedure doitq(stmt varchar2, result in out date)
is
  retval number;
begin
  retval := execq(stmt, result);
end;


procedure trap(
  stmt varchar2
, ok out boolean
, gmsg varchar2 default null
, bmsg varchar2 default null
) is
  t1 date;
  t2 date;
  elapsed varchar2(30);
begin
  t1 := sysdate;
  doit(stmt);
  ok := true;
  t2 := sysdate;
  elapsed := ' ('||ltrim(to_char((t2-t2)*24*60, '990.90'))||'s)';
  if gmsg is not null then
    out(gmsg||elapsed);
  end if;
exception when others then
  ok := false;
  if bmsg is not null then
    out(bmsg);
  end if;
  out(sqlerrm);
end trap;


procedure trap (
  stmt varchar2
, cnt in out number
, gmsg varchar2 default null
, bmsg varchar2 default null
) is
  ok boolean;
begin
  trap(stmt, ok, gmsg, bmsg);
  if ok then
    cnt := nvl(cnt, 0) + 1;
  end if;
end trap;


procedure trap (
  stmt varchar2
, gmsg varchar2 default null
, bmsg varchar2 default null
) is
  ok boolean;
begin
  trap(stmt, ok, gmsg, bmsg);
end trap;


function element (
  str varchar2
, delim varchar2 default ' '
, elem number default 0
, quote varchar2 default null
) return varchar2 is

  result varchar2(2000);

  a binary_integer;
  b binary_integer;
  c binary_integer;
  i binary_integer;
  j binary_integer;
  n binary_integer;

  bq char(1);
  eq char(1);

  function getnext(n binary_integer)
  return number is
    i binary_integer;
    j binary_integer;
    p binary_integer;
    q binary_integer;
  begin
    if bq is null then
      return instr(str,delim,n+1,1);
    end if;
    i:=n+1;
    loop
      j:=instr(str,delim,i,1);
      p:=instr(str,bq,i,1);
      if j<p or p=0 then
	return j;
      else
	q:=instr(str,eq,p+1,1);
	if q=0 then
	  return 0;
	end if;
	i:=q+1;
      end if;
    end loop;
  end;

begin
  if elem < 1 then
    return delim;
  end if;

  /*
  if quote is null then
  */
    -- USE SIMPLE VERSION
    if elem < 1 then
      return delim;
    elsif elem = 1 then
      j := instr(str, delim, 1, elem);
      if j = 0 then
	return str;
      else
	return substr(str, 1, j - 1);
      end if;
    else
      i := instr(str, delim, 1, elem - 1);
      j := instr(str, delim, 1, elem);
      if i = 0 then
	return delim;
      elsif j = 0 then
	return substr(str, i + length(delim));
      else
	return substr(str, i + length(delim), j - i - length(delim));
      end if;
    end if;
/*
  else
    -- USE COMPLEX VERSION
    -- GET QUOTE CHARACTER
    if quote is not null then
      bq := substr(quote, 1, 1);
      if length(quote) > 1 then
	eq := substr(quote, 2, 1);
      else
	eq := bq;
      end if;
    end if;

    -- SKIP n-1 ELEMENTS
    b := 0; n := elem;
    while n > 1 loop
      b := getnext(b);
      if b = 0 then
	return delim;
      end if;
      n := n - 1;
    end loop;

    -- RETURN nth ELEMENT
    a := b;
    b := getnext(a);
    if b>0 then
      result := replace(replace(substr(str, a + 1, b - a - 1), bq), eq);
    else
      result := replace(replace(substr(str, a+1), bq), eq);
    end if;
    return result;
  end if;
*/
end element;


procedure parse_object_name(
  str varchar2
, owner out varchar2
, name out varchar2
, dblink out varchar2
) is
  s varchar2(128);
  i binary_integer;
  j binary_integer;
begin
  s := upper(str);
  i := instr(s, '.');
  j := instr(s, '@');
  dblink := '';
  if i > 0 then
    owner := substr(s, 1, i - 1);
    if j > 0 then
      name := substr(s, i + 1, j - i - 1);
    else
      name := substr(s, i + 1);
    end if;
  else
    owner := user;
    if j > 0 then
      name := substr(s, 1, j - 1);
    else
      name := s;
    end if;
  end if;
  if j > i then
    dblink := substr(s, j + 1);
  end if;
end parse_object_name;


procedure parse_object_name(
  str varchar2
, owner out varchar2
, name out varchar2
) is
  s varchar2(128);
  i binary_integer;
begin
  s := upper(str);
  i := instr(s, '.');
  if i > 0 then
    owner := substr(s, 1, i - 1);
    name := substr(s, i + 1);
  else
    owner := user;
    name := s;
  end if;
end parse_object_name;


function has_sys_privs(unam varchar2, priv varchar2)
return boolean is
  c number;
  cmd varchar2(512);
  flg boolean;
  tmp varchar2(40);
begin
  cmd := 'select privilege from sys.dba_sys_privs p'
    ||' where grantee in ('''||unam||''', ''PUBLIC'')'
    ||' and privilege = '''||priv||'''';
  c := sys.dbms_sql.open_cursor;
  sys.dbms_sql.parse(c, cmd, sys.dbms_sql.native);
  sys.dbms_sql.define_column(c, 1, tmp, 40);
  flg := sys.dbms_sql.execute_and_fetch(c) > 0;
  --sys.dbms_sql.column_value(c, 1, tmp);
  sys.dbms_sql.close_cursor(c);
  return flg;
exception
  when others then
  if sys.dbms_sql.is_open(c) then
    sys.dbms_sql.close_cursor(c);
  end if;
  raise;
end has_sys_privs;


function has_tab_privs(unam varchar2, oown varchar2, onam varchar2
, priv varchar2)
return boolean is
  c number;
  cmd varchar2(512);
  flg boolean;
  tmp varchar2(40);
begin
  cmd := 'select privilege from sys.dba_tab_privs p'
    ||' where grantee in ('''||unam||''',''PUBLIC'')'
    ||' and owner='''||oown||''''
    ||' and table_name='''||onam||''''
    ||' and privilege='''||priv||'''';
  c := sys.dbms_sql.open_cursor;
  sys.dbms_sql.parse(c, cmd, sys.dbms_sql.native);
  sys.dbms_sql.define_column(c, 1, tmp, 40);
  flg := sys.dbms_sql.execute_and_fetch(c) > 0;
  --sys.dbms_sql.column_value(c, 1, tmp);
  sys.dbms_sql.close_cursor(c);
  return flg;
exception
  when others then
  if sys.dbms_sql.is_open(c) then
    sys.dbms_sql.close_cursor(c);
  end if;
  raise;
end has_tab_privs;


function can_access(unam varchar2, oown varchar2, onam varchar2
, olink varchar2 default null, priv varchar2 default 'SELECT')
return boolean is
begin
  if olink is not null then
    raise_application_error(-20000,
      'Currently unable to verify security over database links');
  end if;
  return unam = oown
    or has_sys_privs(unam, priv||' ANY TABLE')
    or has_tab_privs(unam, oown, onam, priv);
end can_access;


function can_select(unam varchar2, oown varchar2, onam varchar2
, olink varchar2 default null)
return boolean is
begin
  return can_access(unam, oown, onam, olink, 'SELECT');
end can_select;


function can_insert(unam varchar2, oown varchar2, onam varchar2
, olink varchar2 default null)
return boolean is
begin
  return can_access(unam, oown, onam, 'INSERT');
end can_insert;


function can_update(unam varchar2, oown varchar2, onam varchar2
, olink varchar2 default null)
return boolean is
begin
  return can_access(unam, oown, onam, 'UPDATE');
end can_update;


function can_delete(unam varchar2, oown varchar2, onam varchar2
, olink varchar2 default null)
return boolean is
begin
  return can_access(unam, oown, onam, 'DELETE');
end can_delete;


-- ************
-- UPDATE_TABLE
-- ************


procedure update_table_help
is
begin
  out(version);
  out('>ABSTRACT:');
  out('>  This procedure will update and/or compare a table with data from another');
  out('>  table with similar structure.  The routine dynamically generates an');
  out('>  anonymous PL/SQL block to perform the update based on the underlying');
  out('>  structure of both tables, a unique key, and parameters you specify.');
  out('>  This technique eliminates the errors caused by lengthy SQL statements,');
  out('>  adjusts itself automatically at runtime if tables structures change,');
  out('>  and allows a great number of configurable parameters to be used--making');
  out('>  the routine extremely flexible--without affecting performance.');
  out('>');
  out('>PARAMETERS:');
  out('>  DTAB         destination [owner.]table (default=null)');
  out('>  STAB         source [owner.]table (default=null)');
  out('>  KEYCOLS      specify uid columns (default=null; will get from PK if null)');
  out('>  FLAGS        operation control flags (default=''ie''; see below)');
  out('>  WHERECLAUSE  filters rows from source to apply (default=null)');
  out('>  EXCLCMP      exclude these columns when testing for update (default=null)');
  out('>  EXCLUPD      exclude these columns when applying an update (default=null)');
  out('>  EXCLINS      exclude these columns when applying an insert (default=null)');
  out('>  RBSEG        specify rollback segment to use (default=null)');
  out('>  LIMTOT       stop after N rows processed (0=unlimited)');
  out('>  LIMUPD       stop after N rows updated/inserted (0=unlimited)');
  out('>  LIMERR       stop after N errors inserting or updating (0=unlimited)');
  out('>  COMMITPOINT  commit after every N rows (0=no commit)');
  out('>  ONCOMMIT     execute this SQL at every commit point (default=null)');
  out('>  SKIPREC      skip N rows in source (default=0)');
  out('>  BTAB         insert bad rows (w/errors) into this table with same structure');
  out('>               as STAB (default=null)');
  out('>  NEXTDATE     specifies job start time when used with "j" flag');
  out('>               (default=null=SYSDATE)');
  out('>  INTERVAL     specified job internal (default=null; see DBMS_JOB for details)');
  out('>  PROCNAME     create or replace code as a new stored procedure');
  out('>');
  out('>NOTES:');
  out('>  Only DTAB and STAB are mandetory (even through they default to NULL).');
  out('>  If DTAB is null then this help page is displayed.  It is also recommended');
  out('>  that the package is created and executed in the same schema as DTAB.');
  out('>  It is easiest to use named notation for listing parameters when invoking');
  out('>  the procedure, for example:');
  out('>');
  out('>  SQL> SET SERVEROUTPUT ON SIZE 10240');
  out('>  SQL> exec lang_updtab.update_table(dtab=>''TABLE1'',stab=>''TABLE2'',...');
  out('>');
  out('>  If you don''t SET SERVEROUTPUT ON (10240 or larger recommended)');
  out('>  before running the procedure, you will not see any output.  A summary');
  out('>  report of important statistics is generated via DBMS_OUTPUT.');
  out('>  If the procedure is as a job (via the "j" flag) the user must COMMIT');
  out('>  before the job will begin executing.  This is the normal behavior of');
  out('>  the DBMS_JOB package.');
  out('>');
  out('>FLAGS:');
  out('>  -GENERAL-');
  out('>  i=insert rows not in dest');
  out('>  a=append rows without checking if already exist (requires "i")');
  out('>  u=update rows which already exist');
  out('>  c=update only if values have changed (requires "u")');
  out('>  d=delete rows not in source (rarely used)');
  out('>');
  out('>  -CODE GENERATION-');
  out('>  o=passes through source in key order (requires sort on source--rarely used)');
  out('>  s=generate set-based pl/sql (default is row-based)');
  out('>  x=use discrete transactions');
  out('>  p=commitpoint based on N rows PROCESSED instead of N rows UPDATED');
  out('>  f=filter source to obey foreign key constraints');
  out('>  n=ignore NULLs in source');
  out('>');
  out('>  -ERROR HANDLING-');
  out('>  e=echo all ORA-NNNNN errors');
  out('>  b=break on error (don''t trap exceptions)');
  out('>  w=display warnings');
  out('>');
  out('>  -REPORTING-');
  out('>  r=report changes only');
  out('>  v=verbose reporting (requires "r")');
  out('>  y=tally differences by column (requires "ucr")');
  out('>');
  out('>  -MISC-');
  out('>  l=send pl/sql to DBMS_OUTPUT instead of executing directly');
  out('>    (view and/or spool the code being generated!)');
  out('>  j=submit as a job (use NEXTDATE and/or INTERVAL parameters)');
  out('>  g=verify invoking user''s access to tables');
end update_table_help;


procedure update_table(
  dtab varchar2 default null
, stab varchar2 default null
, flags varchar2 default 'i'
, keycols varchar2 default null
, whereclause varchar2 default null
, exclcmp varchar2 default null
, exclupd varchar2 default null
, exclins varchar2 default null
, rbseg varchar2 default null
, limtot number default 0
, limupd number default 0
, limerr number default 0
, commitpoint number default 0
, oncommit varchar2 default null
, skiprec number default 0
, exceptions varchar2 default null
, nextdate date default null
, interval varchar2 default null
, procname varchar2 default null
)
is
  -- CONSTANTS
  valid_flags constant varchar2(36) := 'ABCDEF  IJ L NOP RS UVWXY ';
  int_type constant varchar2(30) := 'binary_integer';

  -- VARIABLES
  stmt varchar2(2000);
  i myint;
  j myint;
  c varchar2(30);

  cmd varchar2(32000) := '';
  fke varchar2(2000)  := '';
  tmp varchar2(256)  := '';
  num_cols number;

  sowner varchar2(30);
  downer varchar2(30);
  sname varchar2(30);
  dname varchar2(30);
  sdblink varchar2(128);
  ddblink varchar2(128);
  t1 date;
  t2 date;

  key1 varchar2(256);
  srcetab varchar2(128);
  desttab varchar2(128);
  msg varchar(256);

  type co2rec_t is record (
    name varchar2(30)
  , pos1 number
  , typ1 varchar2(30)
  , len1 number
  , nul1 char(1)
  , key1 number
  , pos2 number
  , typ2 varchar2(30)
  , len2 number
  , nul2 char(1)
  , key2 number
  );
  type co2tab_t is table of co2rec_t
    index by binary_integer;
  coltab co2tab_t;
  colnum binary_integer;

  type keyrec_t is record (
    name varchar2(30)
  );
  type keytab_t is table of keyrec_t
    index by binary_integer;
  keytab keytab_t;
  keynum binary_integer;


  procedure put(s varchar2, ok boolean default true) is
  begin
    if ok then
      if isflag(flags, 'l') then
	out(s);
      else
	cmd := cmd||ltrim(s)||chr(10);
      end if;
    end if;
  end put;


  function key_list(keys varchar2, fmt1 varchar2 default '?',
    fmtn varchar2 default '*') return varchar2 is
  -- Return formatted string based on comma-delimited string of key values
  -- <col_name> is substitued for "?"
  -- <fmt1> is substituted for "*"
    i myint;
    t varchar2(30);
    retval varchar2(240) := '';
  begin
    for i in 1..8
    loop
      t := element(keys,',',i);
      exit when t=',';
      if t is not null then
	if i = 1 then
	  retval := retval||replace(fmt1,'?',t);
	else
	  retval := retval||replace(replace(fmtn,'*',fmt1),'?',t);
	end if;
      end if;
    end loop;
    return retval;
  end key_list;


  procedure init_co2tab
  is
    c co2rec_t;
    k keyrec_t;
    i number;
    j number;
    b boolean;
    cmd varchar2(512);
    sdb varchar2(128);
    c1 number;
    retval number;
  begin
    -- VERIFY DTAB EXISTS
    select count(*) into i
    from all_tables
    where owner = downer
    and table_name = dname;
    if i = 0 then
      raise_application_error(-20000, 'Destination table "'||downer||'.'||dname||'" not found!');
    end if;

    -- LOAD COLUMN DEFINITIONS
    colnum := 0;
    j := 0;

    c1 := dbms_sql.open_cursor;

    if sdblink is not null then
      sdb := '@'||sdblink;
    else
      sdb := '';
    end if;

    cmd :=
    'select '
    ||'  dtc2.column_name name '
    ||', dtc1.column_id pos1 '
    ||', dtc1.data_type typ1 '
    ||', dtc1.data_length len1 '
    ||', dtc1.nullable nul1 '
    ||', dtc2.column_id pos2 '
    ||', dtc2.data_type typ2 '
    ||', dtc2.data_length len2 '
    ||', dtc2.nullable nul2 '
    ||'from all_tab_columns'||sdb||' dtc1, all_tab_columns dtc2 '
    ||'where dtc2.owner = '''||downer||''' '
    ||'and dtc2.table_name = '''||dname||''' '
    ||'and dtc1.owner(+) = '''||sowner||''' '
    ||'and dtc1.table_name(+) = '''||sname||''' '
    ||'and dtc1.column_name(+) = dtc2.column_name '
    ||'order by dtc2.column_name ';

    dbms_sql.parse(c1, cmd, dbms_sql.native);

    dbms_sql.define_column(c1,  1, c.name, 30);
    dbms_sql.define_column(c1,  2, c.pos1);
    dbms_sql.define_column(c1,  3, c.typ1, 15);
    dbms_sql.define_column(c1,  4, c.len1);
    dbms_sql.define_column(c1,  5, c.nul1, 1);
    dbms_sql.define_column(c1,  6, c.pos2);
    dbms_sql.define_column(c1,  7, c.typ2, 15);
    dbms_sql.define_column(c1,  8, c.len2);
    dbms_sql.define_column(c1,  9, c.nul2, 1);

    retval := dbms_sql.execute(c1);

    loop
      if dbms_sql.fetch_rows(c1) = 0 then
	exit;
      end if;

      dbms_sql.column_value(c1,  1, c.name);
      dbms_sql.column_value(c1,  2, c.pos1);
      dbms_sql.column_value(c1,  3, c.typ1);
      dbms_sql.column_value(c1,  4, c.len1);
      dbms_sql.column_value(c1,  5, c.nul1);
      dbms_sql.column_value(c1,  6, c.pos2);
      dbms_sql.column_value(c1,  7, c.typ2);
      dbms_sql.column_value(c1,  8, c.len2);
      dbms_sql.column_value(c1,  9, c.nul2);

      colnum := colnum + 1;
      coltab(colnum) := c;
      if c.typ1 is not null then
	j := j + 1;
      end if;
    end loop;

    -- STOP! NO INTERSECTING COLUMNS FOUND
    if j = 0 then
      raise_application_error(-20000, 'No columns found in source (table may not exist)');
    end if;

    -- LOAD PK DEFINITION
    keynum := 0;
    if keycols is null then
      -- DETERMINE FROM PRIMARY KEY / UNIQUE KEYS
      for r in (
	select owner, constraint_name
	from all_constraints
	where owner = downer
	and table_name = dname
	and constraint_type in ('P','U')
	order by decode(constraint_type, 'P', 1, 2)
      ) loop
	for s in (
	  select column_name
	  from all_cons_columns
	  where owner = r.owner
	  and constraint_name = r.constraint_name
	  order by position
	) loop
	  k.name := s.column_name;
	  keynum := keynum + 1;
	  keytab(keynum) := k;
	end loop;
	exit;
      end loop;
    else
      -- SPECIFIED IN KEYCOLS
      i := 0;
      loop
	i := i + 1;
	k.name := upper(ltrim(rtrim(substr(element(keycols, ',', i), 1, 30))));
	exit when k.name = ',';
	if k.name is not null and length(k.name) > 0 then
	  keynum := keynum + 1;
	  keytab(keynum) := k;
	end if;
      end loop;
    end if;

    -- MUST SPECIFY PK
    if keynum = 0 then
      raise_application_error(-20000, 'No PRIMARY KEY found in destination table');
    end if;

    -- VERIFY KEY COLUMNS IN SOURCE AND DEST
    b := false;
    for j in 1..keynum loop
      for i in 1..colnum loop
	if keytab(j).name = coltab(i).name then
	  if coltab(i).typ1 is null then
	    raise_application_error(-20000, 'Key column '||keytab(j).name||' not found in source');
	  end if;
	  coltab(i).pos2 := keynum;
	  b := true;
	  exit;
	end if;
      end loop;
      if not b then
	raise_application_error(-20000, 'Key column '||keytab(j).name||' not found in destination');
      end if;
    end loop;

    -- SET VALUE OF KEY1
    for i in 1..keynum loop
      if i = 1 then
	key1 := keytab(i).name;
      else
	key1 := key1||','||keytab(i).name;
      end if;
    end loop;
  end init_co2tab;


  procedure put_list(
--    o1 varchar2, t1 varchar2          -- TABLE1
--  , o2 varchar2, t2 varchar2          -- TABLE 2
    ex varchar2 default ''              -- EXCLUDE LIST
  , fmt1 varchar2 default '?'           -- FORMAT 1st ENTRY
  , fmtn varchar2 default ',*'          -- FORMAT ALL OTHER ENTRIES
  , nul1 varchar2 default null
  , nul2 varchar2 default null
	-- null=no skip
	-- N=skip if nullable
	-- Y=skip if not nullable
  ) is
  -- Formatted output based on the intersection of columns in two tables
  -- <col_name> is substitued for "?"
  -- <col_id> is substitued for "?#"
  -- <fmt1> is substituted for "*"
    i number;
    j number;
    k number;
    flag boolean;
    line varchar2(256);
    text varchar2(256);
    excl varchar2(256) := null;
    incl varchar2(256) := null;
    tmp varchar2(256);


    function in_list(list varchar2, str varchar2, delim varchar2 default ',')
    -- RETURN TRUE IF STR IN LIST
    -- SUPPORT % WILDCARD
    return boolean is
      s varchar2(256);
      k pls_integer;
    begin
      if instr(list, delim||str||delim) > 0 then
	return true;
      end if;
      if instr(list, '%') > 0 then
	k := 0;
	loop
	  k := k + 1;
	  s := element(list, delim, k);  -- SUBSTR(1,30)?
	  exit when s = delim;
	  if str like s then
	    return true;
	  end if;
	end loop;
      end if;
      return false;
    end;

  begin
    i := 0;
    line := '';

    -- FORMAT EXCLUDE LIST: REMOVE SPACES AND SURROUND WITH ','
    if substr(ex, 1, 1) = '!'
    then
      incl := ','||upper(replace(substr(ex, 2),' '))||',';
    else
      excl := ','||upper(replace(ex, ' '))||',';
    end if;

    for j in 1..colnum loop
      flag := true;

      if coltab(j).typ1 is not null     -- ALSO IN SOURCE
      and (
	excl is null
	or not in_list(excl, coltab(j).name)
      )
      and (
	incl is null
	or in_list(incl, coltab(j).name)
      )
      and (nul1 is null or nul1 = coltab(j).nul1)
      and (nul2 is null or nul2 = coltab(j).nul2)
      then
	null;
      else
	flag := false;
      end if;

      if flag then
	-- COLUMN IS OK!!!
	i := i + 1;
	if i = 1 then
	  text := fmt1;
	else
	  text := replace(fmtn, '*', fmt1);
	end if;
	text := replace(replace(text,'?#',to_char(coltab(j).pos2)),
	  '?',lower(coltab(j).name));
	put(text);
      end if;

    end loop;
  end put_list;


  function excl_key(excl varchar2, key varchar2 default null)
  -- RETURN EXCLUDE STRING WITH KEY APPENDED
  return varchar2 is
  begin
    if key is null
    then
      return excl;
    else
      if substr(excl, 1, 1) = '!'
      then
	return excl;
      else
	return key||','||excl;
      end if;
    end if;
  end excl_key;


  procedure put_fk_filter(v_prepend varchar2, v_alias varchar2)
  is
    r_owner varchar2(30);
    r_table varchar2(30);
    i binary_integer;
    j binary_integer;
  begin
    -- FOR EACH FK
    i := 0;
    for r in (
      select *
      from all_constraints
      where owner = downer
      and table_name = dname
      and constraint_type = 'R'
    ) loop
      i := i + 1;
      if i = 1 then
	put(v_prepend||' exists (');
      else
	put('and exists (');
      end if;
      -- GET REFERENCED TABLE
      for s in (
	select *
	from all_constraints
	where owner = r.r_owner
	and constraint_name = r.r_constraint_name
      ) loop
	put('select 0 from '||s.owner||'.'||s.table_name);
      end loop;
      -- FOR EACH COLUMN
      j := 0;
      for s in (
	select
	  c2.column_name nam2
	, cr.column_name namr
	from all_cons_columns c2, all_cons_columns cr
	where c2.owner = r.owner
	and c2.constraint_name = r.constraint_name
	and cr.owner = r.r_owner
	and cr.constraint_name = r.r_constraint_name
	and c2.position = cr.position
	order by c2.position
      ) loop
	j := j + 1;
	if j = 1 then
	  put('where '||s.namr||'='||v_alias||'.'||s.nam2);
	else
	  put('and '||s.namr||'='||v_alias||'.'||s.nam2);
	end if;
      end loop;
      put(')');
    end loop;
  end put_fk_filter;


  procedure put_tx(md binary_integer)
  -- mode: 0=before start 1=middle 2=after complete
  -- note: if mode=1 then put_tx does nothing
  is
  begin
    if not isflag(flags, 'r') then
      if commitpoint > 0 then
	if md > 0 then
	  if oncommit is not null then
	    put(oncommit);
	  end if;
	  put('commit;');
	  put('c:=c+1;');
	end if;
	if md < 2 then
	  -- if using DISCRETE TRANSACTION THEN IGNORE ROLLBACK SEGMENT
	  -- DISCREATE TX NOT ALLOWED WITH S FLAG???
	  if isflag(flags, 'x') and not isflag(flags, 's') then
	    put('sys.dbms_transaction.begin_discrete_transaction;');
	  elsif rbseg is not null then
	    put('sys.dbms_transaction.use_rollback_segment('''||rbseg||''');');
	  end if;
	end if;
      elsif rbseg is not null and md = 0 then
	put('sys.dbms_transaction.use_rollback_segment('''||rbseg||''');');
      end if;
    end if;
    put('null;');
  end put_tx;


  procedure put_warn(msg varchar2)
  is
  begin
    if isflag(flags, 'w') then
      put('put(''-- WARNING! '||msg||''');');
    else
      put('-- WARNING! '||msg);
    end if;
  end put_warn;


  function need_c2
  return boolean is
  begin
    return orflag(flags, 'iu') and not orflag(flags, 'as');
  end need_c2;


  function need_c3
  return boolean is
  begin
    return isflag(flags, 'd') and not isflag(flags, 's');
  end need_c3;

begin
  t1 := SYSDATE;

  -- IF DTAB IS NULL THIS DISPLAY HELP PAGE
  if dtab is null then
    update_table_help;
    return;
  end if;

  if stab is null then
    raise_application_error(-20000, 'Must at least specify STAB and DTAB parameters');
  end if;

  -- CHECK FOR ILLEGAL FLAG COMBINATIONS
  if nvl(length(flags),0) = 0 then
    raise_application_error(-20000, 'No flags specified.');
  end if;
/*
  tmp := translate(upper(flags), valid_flags, '');
  out('HELP:'||tmp);
  if nvl(length(tmp), 0) > 0 then
    raise_application_error(-20000, 'Invalid flags: '||tmp);
  end if;
*/
  if not orflag(flags, 'iucwd') then
    raise_application_error(-20000, 'At least on of I,U,C,D, or W must be specified.');
  end if;
  if orflag(flags, 'v') and not isflag(flags, 'r') then
    raise_application_error(-20000, 'Must specify R with the V flag');
  end if;
  if orflag(flags, 'y') and not isflag(flags, 'ruc') then
    raise_application_error(-20000, 'Must specify U, C, and R with the A flag');
  end if;
  if isflag(flags, 'c') and not isflag(flags, 'u') then
    raise_application_error(-20000, 'Must specify U with the C flag');
  end if;
  if isflag(flags, 'ac') then
    raise_application_error(-20000, 'Cannot specify A and U flags together.');
  end if;


  -- IF j FLAG SPECIFIED THEN SUBMIT AS A JOB
  if isflag(flags, 'j') then
    cmd := 'lang_updtab.update_table('
     ||'dtab=>'''||dtab||''''
     ||',stab=>'''||stab||'''';
    if flags is not null then cmd := cmd||',flags=>'''||replace(upper(flags),'J')||''''; end if;
    if keycols is not null then cmd := cmd||',keycols=>'''||keycols||''''; end if;
    if whereclause is not null then cmd := cmd||',whereclause=>'''||whereclause||''''; end if;
    if exclcmp is not null then cmd := cmd||',exclcmp=>'''||exclcmp||''''; end if;
    if exclupd is not null then cmd := cmd||',exclupd=>'''||exclupd||''''; end if;
    if exclins is not null then cmd := cmd||',exclins=>'''||exclins||''''; end if;
    if rbseg is not null then cmd := cmd||',rbseg=>'''||rbseg||''''; end if;
    if limtot != 0 then cmd := cmd||',limtot=>'||to_char(limtot); end if;
    if limupd != 0 then cmd := cmd||',limupd=>'||to_char(limupd); end if;
    if limerr != 0 then cmd := cmd||',limerr=>'||to_char(limerr); end if;
    if commitpoint != 0 then cmd := cmd||',commitpoint=>'||to_char(commitpoint);end if;
    if oncommit is not null then cmd := cmd||',oncommit=>'''||oncommit||''''; end if;
    if skiprec != 0 then cmd := cmd||',skiprec=>'||to_char(skiprec); end if;
    if exceptions is not null then cmd := cmd||',exceptions=>'''||exceptions||''''; end if;
    cmd := cmd||');';
    --out(cmd);
    sys.dbms_job.submit(i, cmd, nvl(nextdate,sysdate), interval);
    out('Job '||to_char(i)||' submitted.');
    return;
  end if;

  -- PARSE TABLE NAMES
  if substr(stab, 1, 1) = '@' then
    parse_object_name(dtab||stab, sowner, sname, sdblink);
  else
    parse_object_name(stab, sowner, sname, sdblink);
  end if;
  parse_object_name(dtab, downer, dname, ddblink);
  if downer = 'SYS' and not isflag(flags, 'r') then
    raise_application_error(-20000, 'Destination table cannot be owned by SYS');
  else
    -- check if view pointing to SYS object??
    null;
  end if;

  -- VERIFY DTAB ACCESS
  if isflag(flags, 'g')
  and (downer != user or sowner != user) then

    -- must be able to access DBA_TAB_PRIVS and DBA_SYS_PRIVS
    select count(*)
    into i
    from all_tab_privs
    where table_schema = 'SYS'
    and table_name in ('DBA_TAB_PRIVS','DBA_SYS_PRIVS');
    if i < 2 then
      raise_application_error(-20000,
	'Package owner must have access to DBA_SYS_PRIVS and DBA_TAB_PRIVS'
	||' to verify security');
    end if;

    -- check access to DTAB
    if not can_select(user, downer, dname)
    or (isflag(flags, 'i') and not can_insert(user, downer, dname))
    or (isflag(flags, 'u') and not can_update(user, downer, dname))
    or (isflag(flags, 'd') and not can_delete(user, downer, dname))
    then
      raise insufficient_privileges;
    end if;

    -- check access to STAB
    if not can_select(user, sowner, sname, sdblink) then
      raise insufficient_privileges;
    end if;

  end if;

  /*
  sys.dbms_utility.name_tokenize(stab, sowner, sname, c1, sdblink, nextpos1);
  if sname is null then sname := sowner; sowner := user; end if;
  sys.dbms_utility.name_tokenize(dtab, downer, dname, c2, ddblink, nextpos2);
  if dname is null then dname := downer; downer := user; end if;
  */
  srcetab:=sowner||'.'||sname;
  if sdblink is not null then srcetab:=srcetab||'@'||sdblink; end if;
  desttab:=downer||'.'||dname;
  if ddblink is not null then
    raise_application_error(-20000, 'Destination table may not be remote');
  --desttab:=desttab||'@'||ddblink;
  end if;

  -- GET COLUMN AND KEY INFORMATION
  init_co2tab;
  out('Unique key is ('||key1||')', not isflag(flags, 'l'));

  -- OUTPUT PROCEDURE HEADER
  if procname is not null then
    put('create or replace procedure '||procname||' as');
  else
    put('declare ');
  end if;

  if need_c2 then
    put('  cursor c2 ('||key_list(key1, 'v_? varchar2 ',',*')||')');
    put('  is ');
    put('    select * from '||desttab||' ');
    put('    where '||key_list(key1,'?=v_? ','and *'));
    put('    for update;');
    put('  r2 c2%rowtype;');
  end if;

  if need_c3 then
    put('cursor c3 is');
    put('select * from '||desttab||' dtab');
    put('where not exists (');
    put('select 0 from srcetab');
    put('where '||key_list(key1,'?=dtab.? ','and *'));
    put(')');
    put('for update');
  end if;

  put('  flag boolean;');
  put('  changed boolean;');
  if isflag(flags, 'y') then
    put('  type t_ccol is table of varchar2(30) index by binary_integer;');
    put('  ccol t_ccol;');
    put('  type t_ccnt is table of number index by binary_integer;');
    put('  ccnt t_ccnt;');
  end if;
  put('  subtype myint is binary_integer;');
  put('  st date;');
  put('  et date;');
  put('  c myint := 0;');       -- commits
  put('  n myint := 0;');       -- total processed
  put('  t myint := 0;');       -- total processed + skipped
  put('  m myint := 0;');       -- exists
  put('  d myint := 0;');       -- deleted
  put('  u myint := 0;');       -- updated
  put('  i myint := 0;');       -- inserted
  put('  ie myint := 0;');      -- insert errors
  put('  ue myint := 0;');      -- update errors
  put('  xx myint;');           -- temporary
  put('  yy myint;');           -- temporary
  put('  msg varchar2(240);');  -- temp err message
  put('  err varchar2(240);');  -- temp err message
  put('  smod varchar2(48);');  -- set application MODULE
  put('  sact varchar2(32);');  -- set application ACTION
  put('  scli varchar2(64);');  -- set application CLIENT_INFO
  put('  procedure put(s varchar2) is');
  put('  begin');
  put('    sys.dbms_output.put_line(s);');
  put('  end put;');

  -- START MAIN BLOCK
  put('begin');
  -- INITIALIZE
  put_tx(0);
  put('  st:=sysdate;');

  put('  sys.dbms_application_info.read_module(smod, sact);');
  put('  sys.dbms_application_info.read_client_info(scli);');
  put('  sys.dbms_application_info.set_module(''UPDATE_TABLE'','''
    ||desttab||'('||flags||')'');');

  put('  put('''||version||''');');
  put('  put(''User is "''||user||''"'');');
  if isflag(flags, 'y') then
    put('for xx in 1..255 loop');
    put('ccol(xx):='''';ccnt(xx):=-1;');
    put('end loop;');
  end if;

  -- Report columns not found in Source table
  num_cols := 0;
  for i in 1..colnum
  loop
    num_cols := greatest(num_cols,coltab(i).pos2);
    if coltab(i).typ1 is not null then
      if isflag(flags, 'y') then
	put('ccol('||to_char(coltab(i).pos2)||'):='''||coltab(i).name||''';');
	put('ccnt('||to_char(coltab(i).pos2)||'):=0;');
      end if;
    end if;

    if coltab(i).typ1 is null then
      msg := 'Column '||coltab(i).name||' not found in SOURCE table';
      if coltab(i).nul2 = 'N' then
	msg := '(02) '||msg||' and is NOT NULL';
      else
	msg := '(01) '||msg;
      end if;
      put_warn(msg);
    else
      if coltab(i).typ2 <> coltab(i).typ1
      and not (coltab(i).typ1 like '%CHAR%' and coltab(i).typ2 like '%CHAR%')
      then
	put_warn('(03) Column '||coltab(i).name||' has different datatypes');
      else
	if coltab(i).typ2 <> coltab(i).typ1
	and (coltab(i).typ1 like '%CHAR%' and coltab(i).typ2 like '%CHAR%') then
	  put_warn('(04) Column '||coltab(i).name||' uses mixed string types');
	end if;
	if coltab(i).nul2 = 'N' and coltab(i).nul1 = 'Y' then
	  put_warn('(05) Column '||coltab(i).name||' is going from NULL to NOT NULL');
	end if;
	if coltab(i).len2 < coltab(i).len1 then
	  put_warn('(06) Column '||coltab(i).name||' is larger in SOURCE');
	end if;
      end if;
    end if;
  end loop;

  if isflag(flags, 'd') then
    put_tx(1);
    if isflag(flags, 's') then
      put('-- DELETE RECORD(S)');
      put('begin');
      if isflag(flags, 'r') then
	put('select count(*) into xx');
	put(' from '||desttab||' stab ');
      else
	put('delete from '||desttab||' stab');
      end if;
      put('where not exists (');
      put('  select 0 from '||srcetab||' ');
      -- put('  where '||key1||'=stab.'||key1||' ');
      put('  where '||key_list(key1,'?=stab.? ','and *'));
      put(')');
      if whereclause is not null then
	put('and '||whereclause);
      end if;
      put(';');
      if isflag(flags, 'r') then
	put('d:=xx;');
      else
	put('d:=sql%rowcount;');
      end if;
      if not isflag(flags, 'b') then
	put('exception when others then');
	if isflag(flags, 'e') then
	  put('put(substr(SQLERRM,1,240));');
	else
	  put('null;');
	end if;
      end if;
      put('end;');
    else
      put('for r3 in c3 loop');
      put('delete from '||desttab);
      put('where current of c3;');
      put('d:=d+1;');
      if commitpoint > 0 then
	if commitpoint > 1 then
	  put('if mod(d,'||to_char(commitpoint)||')=0 then');
	  put_tx(1);
	  put('end if;');
	else
	  put_tx(1);
	end if;
      end if;
      put('end loop;');
      put_tx(1);
    end if;
  end if;

  if orflag(flags, 'iu') then
    if isflag(flags, 's') then
      -- SET LEVEL
      -- Flags supported: DIU RE

      if isflag(flags, 'u') then
	put_tx(1);
	put('-- UPDATING RECORDS');
	put('begin');
	if isflag(flags, 'r') then
	  put('select count(*) into xx');
	  put('from '||desttab||' stab');
	else
	  put('update '||desttab||' stab');
	  put('set (');
	  put_list(excl_key(exclupd,key1),'?',',*');
	  put(')=(');
	  put('select ');
	  if isflag(flags, 'n') then
	    -- IGNORE NULLS IN SOURCE
	    put_list(excl_key(exclupd,key1),'nvl(?,stab.?)',',*');
	  else
	    put_list(excl_key(exclupd,key1),'?',',*');
	  end if;
	  put('from '||srcetab);
	  put('where '||key_list(key1,'?=stab.? ','and *'));
	  put(')');
	end if;
	put('where exists (');
	put('select 0 from '||srcetab);
	put('where '||key_list(key1,'?=stab.? ','and *'));
	if isflag(flags, 'c') then
	  put('and (');
	  if isflag(flags, 'n') then
	    put_list(excl_key(exclcmp,key1),
	      'decode(nvl(?,stab.?),stab.?,1,0)=0','or *');
	  else
	    put_list(excl_key(exclcmp,key1),
	      'decode(?,stab.?,1,0)=0','or *');
	  end if;
	  put(')');
	end if;
	put(')');
	if whereclause is not null then
	  put('and '||whereclause);
	end if;
	put(';');
	if isflag(flags, 'r') then
	  put('u:=xx;');
	else
	  put('u:=sql%rowcount;');
	end if;
	if not isflag(flags, 'b') then
	  put('exception when others then');
	  if isflag(flags, 'e') then
	    put('put(substr(SQLERRM,1,240));');
	  else
	    put('null;');
	  end if;
	end if;
	put('end;');
      end if;

      if isflag(flags, 'i') then
	put_tx(1);
	put('-- INSERTING RECORDS');
	put('begin');
	if not isflag(flags, 'r') then
	  put('insert into '||desttab);
	  put('(');
	  put_list(exclins,'?',',*');
	  put(')');
	end if;
	put('select ');
	if isflag(flags, 'r') then
	  put('count(*) into xx');
	else
	  put_list(exclins,'?',',*');
	end if;
	if not isflag(flags, 'a') then
	  put('from '||srcetab||' stab');
	  put('where not exists (');
	  put('select 0 from '||desttab);
	  put('where '||key_list(key1,'?=stab.? ','and *'));
	  put(')');
	end if;
	if isflag(flags, 'f') then
	  put_fk_filter('and','stab');
	end if;
	if whereclause is not null then
	  put('and '||whereclause);
	end if;
	put(';');
	if isflag(flags, 'r') then
	  put('i:=xx;');
	else
	  put('i:=sql%rowcount;');
	end if;
	if not isflag(flags, 'b') then
	  put('exception when others then');
	  if isflag(flags, 'e') then
	    put('put(substr(SQLERRM,1,240));');
	  else
	    put('null;');
	  end if;
	end if;
	put('end;');
      end if;

    else
      put_tx(1);
      -- RECORD LEVEL
      put('for r1 in (');
      put('  select * from '||srcetab||' stab');
      if isflag(flags, 'f') then
	put_fk_filter('where','stab');
      end if;
      if whereclause is not null then
	if isflag(flags, 'f') then
	  put('and '||whereclause);
	else
	  put('where '||whereclause);
	end if;
      end if;
      if isflag(flags, 'o') then
	put('  order by '||key_list(key1,'?',',*'));
      end if;
      put(') loop');
      put('  t:=t+1;');
      put('  if mod(t,100)=0 then');
      put('    sys.dbms_application_info.set_client_info('
	||'to_char(t)||'' rows processed (''||to_char(sysdate,''Dy HH24:MI:SS'')||'')'');');
      put('  end if;');
      if skiprec > 0 then
	put('  if t>'||to_char(skiprec)||' then');
      end if;
      put('  n:=n+1;');

      if not isflag(flags, 'a') then
	put('  open c2('||key_list(key1,'r1.?',',*')||');');
	put('  fetch c2 into r2;');
	put('  flag:=c2%found;');
	put('  if flag then m:=m+1; end if;');
      else
	put('flag:=false');
      end if;

      if isflag(flags, 'i') then
	put('-- INSERT RECORD(S)');
	put('  if not flag then');
	if isflag(flags, 'r') then
	  if isflag(flags, 'v') then
	    put('put(''Would insert '||key1||': '''||
	      key_list(key1,'||r1.?','||'',''*')||');');
	  end if;
	  put('i:=i+1;');
	else
	  put('  begin ');
	  put('     insert into '||desttab||' (');
	  put_list(exclins);
	  put('     ) values (');
	  put_list(exclins,'r1.?');
	  put('     );');
	  put('     i:=i+1;');
	  if not isflag(flags, 'b') then
	    put('exception when others then');
	    if isflag(flags, 'e') then
	      put('msg:=''Error inserting '||key1||':'''||
		key_list(key1,'||r1.?','||'',''*')||';');
	      put('err:=substr(SQLERRM,1,240);');
	      put('put(msg);');
	      put('put(err);');
	    end if;
	    if exceptions is not null then
	      put('insert into '||exceptions||' (');
	      put_list(null,'?',',*');
	      put(') values (');
	      put_list(null,'r1.?',',*');
	      put(');');
	    end if;
	    put('ie:=ie+1;');
	  end if;
	  put('  end;');
	end if;
	put('  end if;');
      end if;

      if isflag(flags, 'u') then
	put('-- UPDATE RECORD(S)');
	put('  if flag then ');
	if isflag(flags, 'c') then
	  put('changed := false;');
	  put('loop');

	  -- CODE TO BE EXECUTED IF A COLUMN IS CHANGED
	  tmp := 'changed:=true;';
	  if isflag(flags, 'v') then
	    tmp := tmp||'put(''?'');';
	  end if;
	  if isflag(flags, 'y') then
	    tmp := tmp||'ccnt(?#):=ccnt(?#)+1;';
	  else
	    tmp := tmp||'exit;';
	  end if;

	  put_list(excl_key(exclcmp,key1),
	    'if r1.?<>r2.? then '||tmp||' end if;',
	    '*');
	  if not isflag(flags, 'n') then
	    -- IGNORE NULLS IN SOURCE
	    put_list(excl_key(exclcmp,key1),
	      'if r1.? is null and r2.? is not null then '||tmp||' end if;',
	      '*','Y');
	  end if;
	  put_list(excl_key(exclcmp,key1),
	    'if r1.? is not null and r2.? is null then '||tmp||' end if;',
	    '*',null,'Y');
	  put('exit when true;');
	  put('end loop;');
	  put('if changed then');
	end if;

	if isflag(flags, 'r') then
	  if isflag(flags, 'v') then
	    put('put(''Would update '||key1||': '''||
	      key_list(key1,'||r1.?','||'',''*')||');');
	  end if;
	  put('u:=u+1;');
	else
	  put('     begin');
	  put('       update '||desttab||' set ');
	  if isflag(flags, 'n') then
	    put_list(excl_key(exclupd,key1),'?=nvl(r1.?,?)');
	  else
	    put_list(excl_key(exclupd,key1),'?=r1.?');
	  end if;
	  put('       where current of c2;');
	  put('       u:=u+1;');
	  if not isflag(flags, 'b') then
	    put('exception when others then');
	    if isflag(flags, 'e') then
	      put('msg:=''Error updating '||key1||':'''||
		key_list(key1,'||r1.?','||'',''*')||';');
	      put('put(msg);');
	      put('put(substr(SQLERRM,1,240));');
	    end if;
	    if exceptions is not null then
	      put('insert into '||exceptions||' (');
	      put_list(null,'?',',*');
	      put(') values (');
	      put_list(null,'r1.?',',*');
	      put(');');
	    end if;
	    put('ue:=ue+1;');
	  end if;
	  put('     end;');
	end if;

	if isflag(flags, 'c') then
	  put('  end if;');
	end if;

	put('  end if;');
      end if;

      if not isflag(flags, 'a') then
	put('  close c2;');
      end if;

      -- SET LIMITS (SHORTEST FIRST)
      if limerr > 0 then
	put('exit when ie+ue>='||to_char(limerr)||';');
      end if;
      if limupd > 0 then
	put('exit when i+u+ie+ue>='||to_char(limupd)||';');
      end if;
      if limtot > 0 then
	put('exit when n>='||to_char(limtot)||';');
      end if;

      if commitpoint > 0 then
	if commitpoint > 1 then  -- OPTIMIZE IF COMMITPOINT=1
	  if isflag(flags, 'p') then
	    put('if n>0 and mod(n,'||to_char(commitpoint)||')=0 then');
	  else
	    put('if i+u>0 and mod(i+u,'||to_char(commitpoint)||')=0 then');
	  end if;
	end if;
	put_tx(1);
	if commitpoint > 1 then
	  put('end if;');
	end if;
      end if;

      if skiprec > 0 then
	put('end if;');
      end if;
      put('end loop;');
    end if;  -- IFFLAG(s)

    -- put_tx(2);

    if isflag(flags, 'y') then
      put('yy:=0;');
      put('for xx in 1..'||to_char(colnum)||' loop');  -- WAS NUM_COLS
      put('if ccnt(xx)>0 then');
      put('put(to_char(xx)||'':''||ccol(xx)||''=''||to_char(ccnt(xx)));');
      put('yy:=yy+1;');
      put('end if;');
      put('end loop;');
      put('put(to_char(yy)||'' columns modified'');');
    end if;
  end if; -- IFFLAG(i,u)

  put_tx(2);

  put('et:=sysdate;');

  -- CODE FOR EXECUTION REPORT
  put('put(''Updated '||desttab||' <== '||srcetab||
    ' (flags='||upper(flags)||')'');');
  if isflag(flags, 'r') then
    put('put(''(REPORTING ONLY--NO CHANGES MADE!!!)'');');
  end if;

  put('put(to_char(t-n)||'' records skipped'');');
  put('put(to_char(n)||'' records processed'');');
  put('put(to_char(m)||'' records existed'');');
  put('put(to_char(d)||'' records deleted'');');
  put('put(to_char(u)||'' records updated'');');
  put('put(to_char(i)||'' records inserted'');');
  put('put(to_char(c)||'' commits'');');
  put('put(to_char(ue)||'' errors updating'');');
  put('put(to_char(ie)||'' errors inserting'');');
  put('put(''Started:  ''||to_char(st,''Dy Mon DD, YYYY HH24:MI:SS''));');
  put('put(''Finished: ''||to_char(et,''Dy Mon DD, YYYY HH24:MI:SS''));');
  put('put(''Elapsed:  ''||ltrim(to_char((et-st)*(24*60),''9990.99''))||'' mins'');');
  put('if et-st>0 then');
  put('put(to_char(n/((et-st)*24*60*60),''99,990'')||'' recs/s (''||ltrim(to_char(n/((et-st)*24),''9,999,990''))||'' recs/hr)'');');
  put('end if;');
  put('sys.dbms_application_info.set_module(smod, sact);');
  put('sys.dbms_application_info.set_client_info(scli);');

  if need_c2 then
    put('exception when others then ');
    put('if c2%isopen then ');
    put('close c2;');
    put('end if;');
    put('sys.dbms_application_info.set_module(smod, sact);');
    put('sys.dbms_application_info.set_client_info(scli);');
    put('raise;');
  end if;

  -- END PL/SQL BLOCK
  put('end;');

  t2 := SYSDATE;

  if isflag(flags, 'l') then
    put('/');
  else
    doit(cmd);
  end if;
end update_table;


end lang_updtab;
/

show errors package body lang_updtab


