rem History:
rem          16-mar-98  mlang   minor
rem          04-feb-98  mlang   Support @<db> syntax for STAB
rem          04-feb-98  mlang   Update progress in DBMS_APPLICATION_INFO
rem          15-jan-98  mlang   require exec user to have direct/public
rem                             grants to DTAB (may require access to
rem                             DBA_TAB_PRIVS,DBA_SYSPRIVS if not owner)
rem          15-jan-98  mlang   fixed bug with in "include" lists
rem          18-dec-97  mlang   prevent DTAB owner to be 'SYS'
rem          12-dec-97  mlang   enhance support for row delete
rem          12-dec-97  mlang   add support to APPEND insert option
rem          08-dec-97  mlang   add code for parse_object_name()
rem          06-dec-97  mlang   detect UNIQUE keys
rem          01-dec-97  mlang   added "f" flag
rem          01-dec-97  mlang   allow STAB to be a remote table
rem          01-dec-97  mlang   reduce num calls to all_tab_columns
rem          01-dec-97  mlang   added "n" flag
rem          01-dec-97  mlang   add support for wildcards in exclude column lists


create or replace package lang_updtab
as

debugging constant boolean := false;

-- MISC

function version return varchar2;
pragma restrict_references(version, wnds, wnps, rnds, rnps);

function quote(str varchar2) return varchar2;
pragma restrict_references(quote, wnds, wnps, rnds, rnps);

function convert_number(str varchar2, fmt varchar2 default null) return number;
pragma restrict_references(convert_number, wnds, wnps, rnds, rnps);

function convert_date(str varchar2, fmt varchar2 default null) return date;
pragma restrict_references(convert_date, wnds, wnps, rnds, rnps);

function concat_if(str1 varchar2, str2 varchar2, ok boolean default true
, sep varchar2 default null)
return varchar2;
pragma restrict_references(concat_if, wnds, wnps);


-- UPDATE TABLE ROUTINE

procedure update_table(
  dtab varchar2 default null          -- DESTINATION TABLE
, stab varchar2 default null          -- SOURCE TABLE
, flags varchar2 default 'i'          -- CONTROL FLAGS (see below)
, keycols varchar2 default null       -- UNIQUE ID (comma-delimited)
, whereclause varchar2 default null   -- RECORDS TO EXCLUDE FROM SOURCE
, exclcmp varchar2 default null       -- EXCLUDE COLUMNS IN UPDATE/COMPARE
, exclupd varchar2 default null       -- EXCLUDE COLUMNS FROM UPDATE
, exclins varchar2 default null       -- EXCLUDE COLUMNS FROM INSERT
, rbseg varchar2 default null         -- SELECT ROLLBACK SEGMENT
, limtot number default 0             -- STOP WHEN n RECORDS PROCESSED
, limupd number default 0             -- STOP WHEN n RECORDS INS/UPD
, limerr number default 0             -- STOP WHEN n ERRORS OCCURRED
, commitpoint number default 0        -- COMMIT EVERY n RECORDS
, oncommit varchar2 default null      -- EXECUTE SQL ON COMMIT
, skiprec number default 0            -- SKIP N ROWS
, exceptions varchar2 default null    -- REPORT EXCEPTIONS TO THIS TABLE
, nextdate date default null          -- DBMS_JOB.SUBMIT "NEXT_DATE"
, interval varchar2 default null      -- DBMS_JOB.SUBMIT "INTERVAL"
, procname varchar2 default null      -- CREATE OR REPLACE AS "procname"
);


end lang_updtab;
/

show errors package lang_updtab

